"""
Library specific exception definitions.
"""

class r4ven_utilsError(Exception):
    """
    Base r4ven_utils exception that all others inherit.
    This is done to not pollute the built-in exceptions, which *could* result
    in unintended errors being unexpectedly and incorrectly handled within
    implementers code.
    """

class TestFailed(r4ven_utilsError):
    """
    Test failed exception.
    """

class MaxRetriesExceeded(r4ven_utilsError):
    """
    Maximum number of retries exceeded.
    """

class HTMLParseError(r4ven_utilsError):
    """
    HTML could not be parsed.
    """

class ExtractError(r4ven_utilsError):
    """
    Data extraction based exception.
    """


"""
Test script for my logging related functions.
"""

import os
import sys

# Getting the name of the directory where the this file is present.
current_directory = os.path.dirname(os.path.realpath(__file__))

# Getting the parent directory name where the current directory is present.
parent_directory = os.path.dirname(current_directory)

# Adding the parent directory to the sys.path.
sys.path.append(parent_directory)

# Import the main function [function_logger()] from function_logger.py
from r4ven_utils.log4me import function_logger

# Import logging module to get the logging.levels constants.
import logging

def get_logging_levels(file_level: int) -> list:
    """
    Get the list of logging levels that are greater or equal than file_level
    parameter of the function_logger() function.

    Args:
        file_level (int): Logging level that is passed to function_logger()
        as the parameter file_level.

    Returns:
        list: List containing all logging levels names, as strings, that have a priority
        greater or equal to the file_level parameter of the function_logger() function.
    """
    dict_logging_levels = {logging.DEBUG : "DEBUG",
                      logging.INFO : "INFO",
                      logging.WARNING: "WARNING",
                      logging.ERROR : "ERROR",
                      logging.CRITICAL : "CRITICAL"}

    list_logging_levels = []
    for key, value in dict_logging_levels.items():
        if key >= file_level:
            list_logging_levels.append(value)

    return list_logging_levels

def check_file_level(function_name: str, file_level: int) -> None:
    """
    Check if the log file contains the expected logging levels.

    Args:
        function_name (str): The name of the function that's calling the function
        function_logger.
    """
    logs_directory = os.getcwd()
    logs_directory = os.path.join(logs_directory, "logs")
    function_log_path = os.path.join(logs_directory, "logs", f"log4me/{function_name}.log")

    # Get file object reference for the file.
    with open (file = function_log_path, mode = "r") as file:
        # Read the content of the file to string.
        file_to_be_inspected = file.read()

        logging_levels = get_logging_levels(file_level)

        if all(logging_level in file_to_be_inspected for logging_level in logging_levels):
            print(f"For {function_name}: ")
            print("All logging levels with a priority greater or equal" +\
                  f" {file_level} are present in the {function_name}.log file. \n")
        else:
            sys.exit()

def f1() -> None:
    """
    Function that test a specific case of the function_logger(), from src.function_logger.

    Case 01:

        Args:
            > file_level (str): Logging level that will be written in the log file.
                > file_level = None;
                The function should write all logging levels with priority equal or above DEBUG
                to the f1.log file.

            > console_level (str, optional): Logging level that will be displayed
            at the console.
                > console_level = None;
                The function shouldn't write any message to the console, because there isn't any
                logging level assigned to the console_level variable.
    """
    f1_logger = function_logger(file_level = logging.DEBUG)
    f1_logger.debug("f1 Debug message")
    f1_logger.info("f1 Information message")
    f1_logger.warning("f1 Warning message")
    f1_logger.error("f1 Error message")
    f1_logger.critical("f1 Critical message")

def f2() -> None:
    """
    Function that test a specific case of the function_logger(), from src.function_logger.

    Case 02:

        Args:
            > file_level (str): Logging level that will be written in the log file.
                > file_level = logging.WARNING;
                The function should write all logging levels with priority equal or above WARNING
                to the f2.log file.

            > console_level (str, optional): Logging level that will be displayed
            at the console.
                > console_level = None;
                The function should write all logging levels with priority equal or above INFO
                to the console.
    """
    f2_logger = function_logger(file_mode = "a",
                                file_level = logging.WARNING,
                                console_level = logging.INFO)
    f2_logger.debug("f2 Debug message")
    f2_logger.info("f2 Information message")
    f2_logger.warning("f2 Warning message")
    f2_logger.error("f2 Error message")
    f2_logger.critical("f2 Critical message")

def test_function_logger() -> None:
    """
    Test possible scenarios of the logging function [function_logger()] of the log4me package.
    """
    try:
        f1()
        check_file_level("f1", logging.DEBUG)
        logging.shutdown()
    except SystemExit:
        print()
    try:
        f2()
        check_file_level("f2", logging.WARNING)
        logging.shutdown()
    except SystemExit:
        print()

if __name__ == "__main__":
    test_function_logger()



"""
This module contains all logging related functions.
"""

import os
import inspect
import logging
import time
import functools

def function_logger(file_mode: str = "w",
                    file_level: int = logging.INFO,
                    console_level: int = None) -> logging.Logger:
    """
    Creates a logger object specific to the function in which it's called.

    Args:
        file_mode (str): A string that define which mode you want to open the log file.
        Defaults to "w" - Write - Opens a file for writing, creates the file if it does not exist.

        file_level (int): Logging level that will be written in the log file.
        Defaults to logging.INFO.

        console_level (int, optional): Logging level that will be displayed at the console.
        Defaults to None.

    Returns:
        logging.Logger: Logger object of the specific function in which this
        function is called.
    """
    create_logs_folder()
    script_name = os.path.basename(__file__).removesuffix(".py")
    create_script_logs_folder(script_name)

    caller_frame = inspect.stack()[1]

    if caller_frame[3] == "<module>":
        # If the caller is the main module (<module>), get the main function name
        function_name = inspect.getmodulename(caller_frame[1])
    else:
        # Otherwise, get the name of the calling function
        function_name = caller_frame[3]

    logger = logging.getLogger(function_name)

    # Check if handlers are already present and if so, clear them before adding new handlers.
    if (logger.hasHandlers()):
        logger.handlers.clear()

    # By default, logs all messages.
    logger.setLevel(logging.DEBUG)

    if console_level != None:
        # StreamHandler logs to console.
        ch = logging.StreamHandler()
        ch.setLevel(console_level)
        ch_format = logging.Formatter("%(levelname)-8s - %(message)s")
        ch.setFormatter(ch_format)
        logger.addHandler(ch)

    # FileHandler logs to file.
    fh = logging.FileHandler(r"logs/{0}/{1}.log".\
        format(script_name, function_name), mode = file_mode)
    fh.setLevel(file_level)
    fh_format = logging.\
        Formatter("%(asctime)s - %(lineno)d - %(levelname)-8s - %(message)s")
    fh.setFormatter(fh_format)
    logger.addHandler(fh)

    return logger

def create_logs_folder() -> None:
    """
    Check if there's a logs folder (/logs) in the current directory,
    if there isn't, create it.
    """
    project_directory = os.getcwd()
    logs_directory = os.path.join(project_directory, "logs")
    if not os.path.exists(logs_directory):
        os.makedirs(logs_directory)

def create_script_logs_folder(script_name: str) -> None:
    """
    Check if there's a log folder for the script that is running
    in the project directory (/logs/script_name), if there isn't, create it.

    Args:
        script_name (str): The name of the script that's calling the function
        function_logger.
    """
    project_directory = os.getcwd()
    script_logs_directory = os.path.join(project_directory, "logs")
    final_directory = os.path.join(script_logs_directory, script_name)
    if not os.path.exists(final_directory):
        os.makedirs(final_directory)

def calculate_execution_time(logger: logging.Logger):
    """
    Decorator that calculates the execution time of a function and logs this time.

    Args:
        logger (logging.Logger): Logger object returned by the function function_logger.

    Returns:
        function: Decorated function.
    """
    def function_decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            """
            Wrapper function that calculates the execution time of the decorated function and logs it.

            Args:
                *args: Positional arguments of the function.
                **kwargs: Keyword arguments of the function.

            Returns:
                Any: Result of the decorated function.
            """
            start_time = time.time()
            result = function(*args, **kwargs)
            end_time = time.time()
            execution_time = end_time - start_time

            logger.info(f"The function {function.__name__}"+\
                f" was executed in {int(execution_time)} seconds.")

            return result

        return wrapper

    return function_decorator



import setuptools

# Reads the content of your README.md into a variable to be used in the setup below.
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    # Package Information:
    name='r4ven_utils',
    packages=['r4ven_utils'], # Should match the package folder.
    version='0.0.5',
    license='GPLv3',

    # Description Information:
    description='The companion that holds your grab-bag of utility functions and objects',
    long_description=long_description,
    long_description_content_type="text/markdown",

    # Author Information:
    author='Victor Vinci Fantucci',
    author_email='victor.v.fantucci@gmail.com',
    url='https://github.com/VictorFantucci/r4ven_utils_dev',

    # Project URLs:
    project_urls =
    {
        "Bug Tracker": "https://github.com/VictorFantucci/r4ven_utils_dev/issues"
    },

    # Project Requirements:
    install_requires=[],

    # Project Descriptive meta-data:
    keywords=['r4ven_utils', 'utils', 'logs'],

    # Project Classifiers:
    classifiers=
    [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Documentation',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
    ],

    download_url="https://github.com/VictorFantucci/r4ven_utils_dev/archive/refs/tag/0.0.5.tar.gz",
)
